__version__ = "0.3.0"
__license__ = "GNU Lesser General Public License v3.0 (LGPL-3.0)"
__copyright__ = "Copyright (C) 2023-present, MemoKing34 <https://github.com/MemoKing34>"


from . import filters, sync
from .clients import ClientManager, idle
